<?php
/**
 * @file
 * Functions for encypting/decrypting all users.
 */

define('DBEE_ALL_USERS_CRYPT_LIMIT', 15);

/**
 * Encrypts or decrypts all user email addresses.
 *
 * This function encrypts or decrypts all user email addresses from the user
 * table. It is used when the dbee module is installed, enabled or disabled, or
 * when the AES encryption options are updated. it acts as simple callback or
 * batch operation if there are more than 15 users in database.
 *
 * @param $action
 *   A string. Possible values are 'decrypt', 'encrypt', and 'change'. If set to
 *   'change', the $decrypt_params or the $encrypt_params must be provided.
 *
 * @return
 *   A boolean False if $action is not correctly set or NULL.
 */
function dbee_update_crypt_all($action, $upgrade_dbee = FALSE) {
  if (!in_array($action, array('decrypt', 'encrypt', 'change'))) {
    return FALSE;
  }

  // Count how much users will be processed.
  $db_users = dbee_stored_users();
  $n_users = count($db_users);

  if ($n_users <= 0) {
    // No users.
    return array();
  }
  if ($n_users <= DBEE_ALL_USERS_CRYPT_LIMIT) {
    // Only 10 users to crypt. No batch operations.
    $context = array();
    dbee_update_crypt_users_all_batch_proceed($action, $upgrade_dbee, $context);
    $success = $context['finished'] >= 1;
    $results = $context['results'];
    dbee_update_crypt_users_all_batch_finished($success, $results, array());
  }
  else {
    // Batch operation.
    $do_message = (($action == 'encrypt') ? t('encrypted') : (($action == 'decrypt') ? t('decrypted') : t('re-encrypted')));
    $batch = array(
      'title' => t('@crypt all users', array('@crypt' => $do_message)),
      'operations' => array(
        array('dbee_update_crypt_users_all_batch_proceed', array($action, $upgrade_dbee)),
      ),
      'finished' => 'dbee_update_crypt_users_all_batch_finished',
      'file' => drupal_get_path('module', 'dbee') . '/includes/dbee.users.inc',
    );
    batch_set($batch);
    // From Drush, explicitly launch the batch proceess.
    if (drupal_is_cli() && function_exists('drush_main')) {
      $batch =& batch_get();
      $batch['progressive'] = FALSE;
      // Process the batch.
      drush_backend_batch_process();
    }
  }
}

/**
 * Batch process for changing encryption of users (10 users per operation).
 *
 * @param $action
 *   A string. Possible values are 'decrypt', 'encrypt', and 'change'. If set to
 *   'change', the $decrypt_params or the $encrypt_params must be provided.
 * @param $context
 *   An array provided by reference, stores batch procesing informations.
 *
 * @see dbee_update_crypt_all()
 */
function dbee_update_crypt_users_all_batch_proceed($action, $upgrade_dbee, &$context) {
  if (!function_exists('dbee_unstore')) {
    // On batch operation, after the module is disabled, the dbee.module file is
    // not loaded anymore. Wee load it now for decrypting emails after the
    // dbee module is disabled.
    module_load_include('module', 'dbee');
  }
  if (empty($context['sandbox'])) {
    $context['sandbox']['progress'] = 0;
    $all_users = dbee_stored_users();
    $context['sandbox']['max'] = count($all_users);

    // Set parameters for displaying message on finished.
    $context['results']['users_total'] = $context['sandbox']['max'];
    $context['results']['dbee_action'] = $action;
    $context['results']['dbee_upgrade'] = $upgrade_dbee;
  }
  $limit = DBEE_ALL_USERS_CRYPT_LIMIT;
  $db_users = dbee_stored_users(NULL, $context['sandbox']['progress'], $limit);
  foreach ($db_users as $uid => $db_values_object) {
    // Change database encryption for this user.
    if (dbee_update_crypt_user($db_values_object, $action, $upgrade_dbee)) {
      $context['results']['users_updated'][] = $uid;
    }

    $context['sandbox']['progress']++;
    $context['message'] = t('processing user : %name (id : @id)', array('@id' => $uid, '%name' => $db_values_object->name));
  }
  if ($context['sandbox']['progress'] < $context['sandbox']['max']) {
    $context['finished'] = $context['sandbox']['progress'] / $context['sandbox']['max'];
  }
  else {
    $context['finished'] = 1;
  }
}

/**
 * Batch process finished callback for changing encryption on all users.
 *
 * @see dbee_update_crypt_all()
 */
function dbee_update_crypt_users_all_batch_finished($success, $results, $operations) {

  $n_updated = (isset($results['users_updated'])) ? count($results['users_updated']) : 0;
  $n_users = (isset($results['users_total'])) ? $results['users_total'] : 0;
  $action = (isset($results['dbee_action'])) ? $results['dbee_action'] : FALSE;
  $upgrade_dbee = (isset($results['dbee_upgrade'])) ? $results['dbee_upgrade'] : FALSE;
  if ($action) {
    $do_message = (($action == 'encrypt') ? t('encrypted') : (($action == 'decrypt') ? t('decrypted') : t('re-encrypted')));
    $message_arg = array(
      '%crypted' => $do_message,
      '@updated_users' => $n_updated,
      '@total_users' => $n_users,
    );
    if ($success) {
      $message_str = 'All users email addresses have been %crypted (concerning @updated_users of @total_users users)';
      $message_status = TRUE;
    }
    else {
      $message_str = 'Failing on the operation : users email addresses have been %crypted (concerning @updated_users of @total_users users)';
      $message_status = FALSE;
    }
    // Watchdog changes.
    $wd_status = ($message_status) ? WATCHDOG_INFO : WATCHDOG_ERROR;
    watchdog('dbee', $message_str, $message_arg, $wd_status);
    $mess_status = ($message_status) ? 'status' : 'error';
    drupal_set_message(t($message_str, $message_arg), $mess_status);

    if ($upgrade_dbee && $n_updated == 0 && $n_users > 0) {
      throw new DrupalUpdateException('The users email could not be decrypted !');
    }
  }
}

/**
 * Encrypts or decrypts one user email addresses.
 *
 * @param $db_values
 *   An object of user database stored values. Keys are 'uid', 'mail', 'init'.
 * @param $action
 *   A string. Possible values are 'decrypt', 'encrypt', and 'change'. If set to
 *   'change', the $decrypt_params or the $encrypt_params must be provided.
 *
 * @return
 *   A boolean, TRUE if the user has been updated.
 */
function dbee_update_crypt_user($db_values, $action, $upgrade_dbee = FALSE) {

  switch ($action) {
    case 'encrypt':
      // Encrypt every email addresses.
      $to_update = dbee_store($db_values);
      break;
    case 'decrypt':
      // Decrypt every email addresses.
      $to_update = dbee_unstore($db_values, $upgrade_dbee);
      break;
    case 'change':
      // First decrypt every email addresses.
      $uncrypted_to_update = dbee_unstore($db_values, $upgrade_dbee);
      // Now encrypt every email addresses with the new parameters.
      $to_update = dbee_store($db_values, $uncrypted_to_update);
      break;
    default:
      // This action is not allowed.
      return FALSE;
  }

  $changes = array();
  $updated = FALSE;
  foreach (array('mail', 'init', 'dbee_mail', 'dbee_init') as $dbee_field) {

    $stored = $db_values->$dbee_field;
    if (isset($to_update[$dbee_field]) && $to_update[$dbee_field] != $stored) {
      $changes[$dbee_field] = $to_update[$dbee_field];
      // Claim that a user as changed only if it has been encrypted (not for
      // erasing a dbee_* field neither for storing lowercase invalid email).
      if (in_array($dbee_field, array('mail', 'init'))) {
        $updated = TRUE;
      }
    }
  }

  $uid = (!empty($db_values->uid)) ? $db_values->uid : FALSE;
  if ($uid && !empty($changes)) {
    // Processed to the update of the users table.
    db_update('users')
      ->fields($changes)
      ->condition('uid', $uid)
      ->execute();
  }

  return $updated;
}

/*
 * Analyse all users encryption
 *
 * Result displayed on the status page (admin/reports/status) and used to stop
 * disabling a module that may be need by the dbee (on admin/modules page).
 *
 * @return
 *   A multidimensional array, keys are 'total_users', 'empty_users',
 *   'invalid_users', 'crypted_users', 'mail_users' and values are the number of
 *   concerned users; and a last key : 'methods' : each key are an encryption
 *   method and value are the number of users using this encryption method.
 */
function dbee_analyse_users() {
  $users = dbee_stored_users();
  $results = array(
    'total_users' => 0,
    'empty_users' => 0,
    'invalid_users' => 0,
    'crypted_users' => 0,
    'mail_users' => 0,
    'methods' => array(),
  );
  foreach ($users as $uid => $account) {
    // We don't care about the init value.
    foreach (array('mail') as $field) {
      $value = $account->$field;
      if (dbee_email_to_alter($value, FALSE)) {
        // This is a not empty string.
        $method = dbee_decrypt_method($value);
        if (!$method) {
          // The data is not encrypted.
          // Test if it should be.
          if (dbee_email_to_alter($value)) {
            // Error, this data should be encrypted.
            $results['mail_users']++;
          }
          else {
            // Invalid email address.
            $results['invalid_users']++;
          }
        }
        else {
          // The data is encrypted.
          $results['crypted_users']++;
          if (!isset($results['methods'][$method])) {
            $results['methods'][$method] = 0;
          }
          $results['methods'][$method]++;
        }
      }
      else {
        // Empty value
        $results['empty_users']++;
      }
    }
    $results['total_users']++;
  }
  // Sort by the most used encryption method.
  arsort($results['methods']);
  return $results;
}

/*
 * Prevent for disabling the real_aes module if needed by the dbee module.
 *
 * Dbee module may use extra module for the encryption method but not have a
 * dependcy to them : the real_aes module. We avoid to disabled such module
 *
 * @see system_modules_submit()
 */
function dbee_form_system_modules_validate($form, &$form_state) {
  // Look for : modulename => encryption method
  $dependency = array('real_aes' => 'authenticated_aes');

  // Builds list of modules.
  $modules = array();
  // If we're not coming from the confirmation form, build the list of modules.
  if (empty($form_state['storage'])) {
    // If we're not coming from the confirmation form, build the module list.
    foreach ($form_state['values']['modules'] as $group_name => $group) {
      foreach ($group as $module => $enabled) {
        $modules[$module] = array('group' => $group_name, 'enabled' => $enabled['enable']);
      }
    }
  }
  else {
    // If we are coming from the confirmation form, fetch
    // the modules out of $form_state.
    $modules = $form_state['storage']['modules'];
  }

  foreach ($dependency as $module => $method) {
    if (isset($modules[$module])) {
      $already_enabled = module_exists($module);
      $disabling = ($already_enabled && !$modules[$module]['enabled']);
      if ($disabling) {
        // Look if the method is used.
        $analyse = dbee_analyse_users();
        if (!empty($analyse['methods'][$method])) {
          $message = t('The %module module can not be disabled because it used by the Database Email Encryption module.', array('%module' => $module));
          $message .= ' ' . t('You need to change first the encryption method on the <a href="!url">dbee admin page</a>', array('!url' => url(ENCRYPT_MENU_PATH . '/edit/dbee')));
          $element_name = "modules][{$modules[$module]['group']}][{$module}][enable";
          // @TODO : element name does not work. Not a big deal.
          // Prevent from disabling the module.
          form_set_error($element_name, $message);
        }
      }
    }
  }
  // On upgrading from prior version to v7.x-3.x : Prevent from disabling aes or
  // dbee module untill the update script has not been run. It could either the
  // aes module or the submodule aes from the real_aes module.
  $undisabling_modules = array(
    'aes' => 'AES',
    'dbee' => 'DataBase Email Encryption',
  );
  foreach($undisabling_modules as $undisabling_module => $module_name) {
    if (isset($modules[$undisabling_module])) {
      // The module is installed
      if (!$modules[$undisabling_module]['enabled'] && module_exists($undisabling_module)) {
        // The module is going to be disabled.
        // Retreive de current update number.
        $db_number = db_select('system', 's')
          ->fields('s', array('schema_version'))
          ->condition('s.name', 'dbee')
          ->condition('s.type', 'module')
          ->execute()->fetchField();
        if (is_numeric($db_number) && $db_number > 0 && $db_number < 7300) {
          $message = t('The %module module can not be disabled.', array('%module' => $module_name));
          $message .= ' <strong>' . t('You need to run <a href="!url">the update script</a> first !', array('!url' => url('update.php', array('absolute' => TRUE)))) . '</strong>';
          $element_name = "modules][{$modules[$undisabling_module]['group']}][{$undisabling_module}][enable";
          // @TODO : element name does not work. Not a big deal.
          // Prevent from disabling the module.
          form_set_error($element_name, $message);
        }
        elseif ($undisabling_module == 'aes') {
          drupal_set_message(t('Information : the <em>AES</em> module is not used anymore by the <em>Database Email Encryption</em> module and can be disabled'));
        }
      }
    }
  }
}
